
from ..abi_contract_wrapper import ABIContractWrapper
from ..solidity_types import *
from ..credentials import Credentials

CONTRACT_ADDRESS =     {
    "cv": "0x75d8ba2E4725633FcdcC165332dCA04c107915cA",
    "sd": "0xeb9ff38209dCC4236DBFb3C275c0AAeEBf0B92Cf"
}

ABI = """[
    {"name": "DarkSummoningStorageSet", "type": "event", "inputs": [{"name": "heroId", "type": "uint256", "internalType": "uint256", "indexed": false}, {"name": "values", "type": "tuple", "internalType": "struct DarkSummoning", "components": [{"name": "levelCarryover", "type": "uint32", "internalType": "uint32"}], "indexed": false}], "anonymous": false},
    {"name": "Gen0RerollSDStorageSet", "type": "event", "inputs": [{"name": "heroId", "type": "uint256", "internalType": "uint256", "indexed": false}, {"name": "values", "type": "tuple", "internalType": "struct Gen0RerollSD", "components": [{"name": "player", "type": "address", "internalType": "address"}, {"name": "summonRecord", "type": "uint32", "internalType": "uint32"}, {"name": "rerolledGameGenes", "type": "bool", "internalType": "bool"}, {"name": "rerolledAppearanceGenes", "type": "bool", "internalType": "bool"}], "indexed": false}], "anonymous": false},
    {"name": "OwnershipTransferred", "type": "event", "inputs": [{"name": "previousOwner", "type": "address", "internalType": "address", "indexed": true}, {"name": "newOwner", "type": "address", "internalType": "address", "indexed": true}], "anonymous": false},
    {"name": "owner", "type": "function", "inputs": [], "outputs": [{"name": "owner_", "type": "address", "internalType": "address"}], "stateMutability": "view"},
    {"name": "transferOwnership", "type": "function", "inputs": [{"name": "_newOwner", "type": "address", "internalType": "address"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "setModerator", "type": "function", "inputs": [{"name": "_address", "type": "address", "internalType": "address"}, {"name": "_access", "type": "bool", "internalType": "bool"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "setUpdater", "type": "function", "inputs": [{"name": "_address", "type": "address", "internalType": "address"}, {"name": "_access", "type": "bool", "internalType": "bool"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "getDarkSummoningStorage", "type": "function", "inputs": [{"name": "_heroId", "type": "uint256", "internalType": "uint256"}], "outputs": [{"name": "", "type": "tuple", "internalType": "struct DarkSummoning", "components": [{"name": "levelCarryover", "type": "uint32", "internalType": "uint32"}]}], "stateMutability": "view"},
    {"name": "setLevelCarryover", "type": "function", "inputs": [{"name": "_heroId", "type": "uint256", "internalType": "uint256"}, {"name": "_value", "type": "uint32", "internalType": "uint32"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "getGen0RerollSDStorage", "type": "function", "inputs": [{"name": "_heroId", "type": "uint256", "internalType": "uint256"}], "outputs": [{"name": "", "type": "tuple", "internalType": "struct Gen0RerollSD", "components": [{"name": "player", "type": "address", "internalType": "address"}, {"name": "summonRecord", "type": "uint32", "internalType": "uint32"}, {"name": "rerolledGameGenes", "type": "bool", "internalType": "bool"}, {"name": "rerolledAppearanceGenes", "type": "bool", "internalType": "bool"}]}], "stateMutability": "view"},
    {"name": "setGen0RerollSDStorage", "type": "function", "inputs": [{"name": "_heroId", "type": "uint256", "internalType": "uint256"}, {"name": "_reroll", "type": "tuple", "internalType": "struct Gen0RerollSD", "components": [{"name": "player", "type": "address", "internalType": "address"}, {"name": "summonRecord", "type": "uint32", "internalType": "uint32"}, {"name": "rerolledGameGenes", "type": "bool", "internalType": "bool"}, {"name": "rerolledAppearanceGenes", "type": "bool", "internalType": "bool"}]}], "outputs": [], "stateMutability": "nonpayable"}
]
"""     

class FlagStorageV2(ABIContractWrapper):
    def __init__(self, chain_key:str, rpc:str):
        contract_address = CONTRACT_ADDRESS[chain_key]
        super().__init__(contract_address=contract_address, abi=ABI, rpc=rpc)

    def owner(self, block_identifier:BlockIdentifier = 'latest') -> address:
        return self.contract.functions.owner().call(block_identifier=block_identifier)

    def transfer_ownership(self, cred:Credentials, _new_owner:address) -> TxReceipt:
        tx = self.contract.functions.transferOwnership(_new_owner)
        return self.send_transaction(tx, cred)

    def set_moderator(self, cred:Credentials, _address:address, _access:bool) -> TxReceipt:
        tx = self.contract.functions.setModerator(_address, _access)
        return self.send_transaction(tx, cred)

    def set_updater(self, cred:Credentials, _address:address, _access:bool) -> TxReceipt:
        tx = self.contract.functions.setUpdater(_address, _access)
        return self.send_transaction(tx, cred)

    def get_dark_summoning_storage(self, _hero_id:uint256, block_identifier:BlockIdentifier = 'latest') -> tuple:
        return self.contract.functions.getDarkSummoningStorage(_hero_id).call(block_identifier=block_identifier)

    def set_level_carryover(self, cred:Credentials, _hero_id:uint256, _value:uint32) -> TxReceipt:
        tx = self.contract.functions.setLevelCarryover(_hero_id, _value)
        return self.send_transaction(tx, cred)

    def get_gen0_reroll_sd_storage(self, _hero_id:uint256, block_identifier:BlockIdentifier = 'latest') -> tuple:
        return self.contract.functions.getGen0RerollSDStorage(_hero_id).call(block_identifier=block_identifier)

    def set_gen0_reroll_sd_storage(self, cred:Credentials, _hero_id:uint256, _reroll:tuple) -> TxReceipt:
        tx = self.contract.functions.setGen0RerollSDStorage(_hero_id, _reroll)
        return self.send_transaction(tx, cred)



from ..abi_contract_wrapper import ABIContractWrapper
from ..solidity_types import *
from ..credentials import Credentials

CONTRACT_ADDRESS =     {
    "cv": "0x6a56222A67df18FC282CD58dCDF12e61Be812f97",
    "sd": "0x0000000000000000000000000000000000000000"
}

ABI = """[
    {"name": "Initialized", "type": "event", "inputs": [{"name": "version", "type": "uint8", "indexed": false, "internalType": "uint8"}], "anonymous": false},
    {"name": "Paused", "type": "event", "inputs": [{"name": "account", "type": "address", "indexed": false, "internalType": "address"}], "anonymous": false},
    {"name": "RaffleClosed", "type": "event", "inputs": [{"name": "raffleId", "type": "uint256", "indexed": true, "internalType": "uint256"}], "anonymous": false},
    {"name": "RaffleDrawn", "type": "event", "inputs": [{"name": "raffleId", "type": "uint256", "indexed": true, "internalType": "uint256"}, {"name": "winner", "type": "address", "indexed": true, "internalType": "address"}], "anonymous": false},
    {"name": "RaffleEntered", "type": "event", "inputs": [{"name": "user", "type": "address", "indexed": true, "internalType": "address"}, {"name": "raffleId", "type": "uint256", "indexed": true, "internalType": "uint256"}, {"name": "tickets", "type": "uint256", "indexed": false, "internalType": "uint256"}], "anonymous": false},
    {"name": "RaffleStarted", "type": "event", "inputs": [{"name": "raffleId", "type": "uint256", "indexed": true, "internalType": "uint256"}, {"name": "raffleType", "type": "tuple", "components": [{"name": "id", "type": "uint256", "internalType": "uint256"}, {"name": "rewards", "type": "address[]", "internalType": "address[]"}, {"name": "rewardAmounts", "type": "uint256[]", "internalType": "uint256[]"}, {"name": "maxWinners", "type": "uint256", "internalType": "uint256"}, {"name": "duration", "type": "uint64", "internalType": "uint64"}], "indexed": false, "internalType": "struct RaffleMaster.RaffleType"}], "anonymous": false},
    {"name": "RaffleTypeActivated", "type": "event", "inputs": [{"name": "raffleTypeId", "type": "uint256", "indexed": true, "internalType": "uint256"}, {"name": "bucket", "type": "uint256", "indexed": false, "internalType": "uint256"}], "anonymous": false},
    {"name": "RaffleTypeAdded", "type": "event", "inputs": [{"name": "id", "type": "uint256", "indexed": true, "internalType": "uint256"}, {"name": "raffleType", "type": "tuple", "components": [{"name": "id", "type": "uint256", "internalType": "uint256"}, {"name": "rewards", "type": "address[]", "internalType": "address[]"}, {"name": "rewardAmounts", "type": "uint256[]", "internalType": "uint256[]"}, {"name": "maxWinners", "type": "uint256", "internalType": "uint256"}, {"name": "duration", "type": "uint64", "internalType": "uint64"}], "indexed": false, "internalType": "struct RaffleMaster.RaffleType"}], "anonymous": false},
    {"name": "RaffleTypeDeactivated", "type": "event", "inputs": [{"name": "raffleTypeId", "type": "uint256", "indexed": true, "internalType": "uint256"}, {"name": "bucket", "type": "uint256", "indexed": false, "internalType": "uint256"}], "anonymous": false},
    {"name": "Unpaused", "type": "event", "inputs": [{"name": "account", "type": "address", "indexed": false, "internalType": "address"}], "anonymous": false},
    {"name": "activeRaffleTypes", "type": "function", "inputs": [{"name": "", "type": "uint256", "internalType": "uint256"}, {"name": "", "type": "uint256", "internalType": "uint256"}], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "addActiveRaffleType", "type": "function", "inputs": [{"name": "_id", "type": "uint256", "internalType": "uint256"}, {"name": "_bucket", "type": "uint256", "internalType": "uint256"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "addAndActivateRaffleType", "type": "function", "inputs": [{"name": "_rewards", "type": "address[]", "internalType": "address[]"}, {"name": "_rewardAmounts", "type": "uint256[]", "internalType": "uint256[]"}, {"name": "_maxWinners", "type": "uint256", "internalType": "uint256"}, {"name": "_duration", "type": "uint64", "internalType": "uint64"}, {"name": "_buckets", "type": "uint256[]", "internalType": "uint256[]"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "addRaffleType", "type": "function", "inputs": [{"name": "_rewards", "type": "address[]", "internalType": "address[]"}, {"name": "_rewardAmounts", "type": "uint256[]", "internalType": "uint256[]"}, {"name": "_maxWinners", "type": "uint256", "internalType": "uint256"}, {"name": "_duration", "type": "uint64", "internalType": "uint64"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "closeRaffles", "type": "function", "inputs": [], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "createRaffle", "type": "function", "inputs": [{"name": "_raffleTypeId", "type": "uint256", "internalType": "uint256"}, {"name": "_bucket", "type": "uint256", "internalType": "uint256"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "currentRaffleBuckets", "type": "function", "inputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "drawWinners", "type": "function", "inputs": [], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "enterRaffle", "type": "function", "inputs": [{"name": "_raffleId", "type": "uint256", "internalType": "uint256"}, {"name": "_tickets", "type": "uint256", "internalType": "uint256"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "entries", "type": "function", "inputs": [{"name": "", "type": "uint256", "internalType": "uint256"}, {"name": "", "type": "uint256", "internalType": "uint256"}], "outputs": [{"name": "", "type": "address", "internalType": "address"}], "stateMutability": "view"},
    {"name": "getCurrentRaffleBuckets", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "uint256[]", "internalType": "uint256[]"}], "stateMutability": "view"},
    {"name": "getCurrentRaffleData", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "tuple[]", "components": [{"name": "id", "type": "uint256", "internalType": "uint256"}, {"name": "raffleType", "type": "uint256", "internalType": "uint256"}, {"name": "startTime", "type": "uint256", "internalType": "uint256"}, {"name": "totalEntries", "type": "uint256", "internalType": "uint256"}, {"name": "endTime", "type": "uint256", "internalType": "uint256"}, {"name": "closedBlock", "type": "uint256", "internalType": "uint256"}, {"name": "winners", "type": "address[]", "internalType": "address[]"}, {"name": "status", "type": "uint8", "internalType": "uint8"}], "internalType": "struct RaffleMaster.Raffle[]"}, {"name": "", "type": "tuple[]", "components": [{"name": "id", "type": "uint256", "internalType": "uint256"}, {"name": "rewards", "type": "address[]", "internalType": "address[]"}, {"name": "rewardAmounts", "type": "uint256[]", "internalType": "uint256[]"}, {"name": "maxWinners", "type": "uint256", "internalType": "uint256"}, {"name": "duration", "type": "uint64", "internalType": "uint64"}], "internalType": "struct RaffleMaster.RaffleType[]"}, {"name": "", "type": "uint256[]", "internalType": "uint256[]"}, {"name": "", "type": "uint256[]", "internalType": "uint256[]"}], "stateMutability": "view"},
    {"name": "getLastRaffleBuckets", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "uint256[]", "internalType": "uint256[]"}], "stateMutability": "view"},
    {"name": "getPreviousRaffleData", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "tuple[]", "components": [{"name": "id", "type": "uint256", "internalType": "uint256"}, {"name": "raffleType", "type": "uint256", "internalType": "uint256"}, {"name": "startTime", "type": "uint256", "internalType": "uint256"}, {"name": "totalEntries", "type": "uint256", "internalType": "uint256"}, {"name": "endTime", "type": "uint256", "internalType": "uint256"}, {"name": "closedBlock", "type": "uint256", "internalType": "uint256"}, {"name": "winners", "type": "address[]", "internalType": "address[]"}, {"name": "status", "type": "uint8", "internalType": "uint8"}], "internalType": "struct RaffleMaster.Raffle[]"}, {"name": "", "type": "tuple[]", "components": [{"name": "id", "type": "uint256", "internalType": "uint256"}, {"name": "rewards", "type": "address[]", "internalType": "address[]"}, {"name": "rewardAmounts", "type": "uint256[]", "internalType": "uint256[]"}, {"name": "maxWinners", "type": "uint256", "internalType": "uint256"}, {"name": "duration", "type": "uint64", "internalType": "uint64"}], "internalType": "struct RaffleMaster.RaffleType[]"}, {"name": "", "type": "uint256[]", "internalType": "uint256[]"}, {"name": "", "type": "uint256[]", "internalType": "uint256[]"}], "stateMutability": "view"},
    {"name": "getRaffleList", "type": "function", "inputs": [{"name": "_raffleBuckets", "type": "uint256[]", "internalType": "uint256[]"}], "outputs": [{"name": "", "type": "tuple[]", "components": [{"name": "id", "type": "uint256", "internalType": "uint256"}, {"name": "raffleType", "type": "uint256", "internalType": "uint256"}, {"name": "startTime", "type": "uint256", "internalType": "uint256"}, {"name": "totalEntries", "type": "uint256", "internalType": "uint256"}, {"name": "endTime", "type": "uint256", "internalType": "uint256"}, {"name": "closedBlock", "type": "uint256", "internalType": "uint256"}, {"name": "winners", "type": "address[]", "internalType": "address[]"}, {"name": "status", "type": "uint8", "internalType": "uint8"}], "internalType": "struct RaffleMaster.Raffle[]"}], "stateMutability": "view"},
    {"name": "getRaffleTicketsAllowanceList", "type": "function", "inputs": [{"name": "_raffleBuckets", "type": "uint256[]", "internalType": "uint256[]"}], "outputs": [{"name": "", "type": "uint256[]", "internalType": "uint256[]"}], "stateMutability": "view"},
    {"name": "getRaffleTicketsList", "type": "function", "inputs": [{"name": "_raffleBuckets", "type": "uint256[]", "internalType": "uint256[]"}], "outputs": [{"name": "", "type": "uint256[]", "internalType": "uint256[]"}], "stateMutability": "view"},
    {"name": "getRaffleTypesList", "type": "function", "inputs": [{"name": "_raffleBuckets", "type": "uint256[]", "internalType": "uint256[]"}], "outputs": [{"name": "", "type": "tuple[]", "components": [{"name": "id", "type": "uint256", "internalType": "uint256"}, {"name": "rewards", "type": "address[]", "internalType": "address[]"}, {"name": "rewardAmounts", "type": "uint256[]", "internalType": "uint256[]"}, {"name": "maxWinners", "type": "uint256", "internalType": "uint256"}, {"name": "duration", "type": "uint64", "internalType": "uint64"}], "internalType": "struct RaffleMaster.RaffleType[]"}], "stateMutability": "view"},
    {"name": "getTicketAllowance", "type": "function", "inputs": [{"name": "_user", "type": "address", "internalType": "address"}, {"name": "_raffleId", "type": "uint256", "internalType": "uint256"}], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "initRaffleBuckets", "type": "function", "inputs": [{"name": "_buckets", "type": "uint256", "internalType": "uint256"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "initialize", "type": "function", "inputs": [{"name": "_raffleTickets", "type": "address", "internalType": "address"}, {"name": "_airdropClaim", "type": "address", "internalType": "address"}, {"name": "_itemMinter", "type": "address", "internalType": "address"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "lastRaffleBuckets", "type": "function", "inputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "pause", "type": "function", "inputs": [], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "paused", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "bool", "internalType": "bool"}], "stateMutability": "view"},
    {"name": "playerEntries", "type": "function", "inputs": [{"name": "", "type": "uint256", "internalType": "uint256"}, {"name": "", "type": "address", "internalType": "address"}], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "raffleTypes", "type": "function", "inputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "outputs": [{"name": "id", "type": "uint256", "internalType": "uint256"}, {"name": "maxWinners", "type": "uint256", "internalType": "uint256"}, {"name": "duration", "type": "uint64", "internalType": "uint64"}], "stateMutability": "view"},
    {"name": "raffles", "type": "function", "inputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "outputs": [{"name": "id", "type": "uint256", "internalType": "uint256"}, {"name": "raffleType", "type": "uint256", "internalType": "uint256"}, {"name": "startTime", "type": "uint256", "internalType": "uint256"}, {"name": "totalEntries", "type": "uint256", "internalType": "uint256"}, {"name": "endTime", "type": "uint256", "internalType": "uint256"}, {"name": "closedBlock", "type": "uint256", "internalType": "uint256"}, {"name": "status", "type": "uint8", "internalType": "uint8"}], "stateMutability": "view"},
    {"name": "removeActiveRaffleType", "type": "function", "inputs": [{"name": "_index", "type": "uint256", "internalType": "uint256"}, {"name": "_bucket", "type": "uint256", "internalType": "uint256"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "setRaffleType", "type": "function", "inputs": [{"name": "_raffleTypeId", "type": "uint256", "internalType": "uint256"}, {"name": "_rewards", "type": "address[]", "internalType": "address[]"}, {"name": "_rewardAmounts", "type": "uint256[]", "internalType": "uint256[]"}, {"name": "_maxWinners", "type": "uint256", "internalType": "uint256"}, {"name": "_duration", "type": "uint64", "internalType": "uint64"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "startRandomRaffle", "type": "function", "inputs": [{"name": "_bucket", "type": "uint256", "internalType": "uint256"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "startRandomRaffles", "type": "function", "inputs": [{"name": "_buckets", "type": "uint256[]", "internalType": "uint256[]"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "timePerTicket", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "totalRaffleTypes", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "totalRaffles", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "unpause", "type": "function", "inputs": [], "outputs": [], "stateMutability": "nonpayable"}
]
"""     

class RaffleMaster(ABIContractWrapper):
    def __init__(self, chain_key:str, rpc:str):
        contract_address = CONTRACT_ADDRESS[chain_key]
        super().__init__(contract_address=contract_address, abi=ABI, rpc=rpc)

    def active_raffle_types(self, a:uint256, b:uint256, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.activeRaffleTypes(a, b).call(block_identifier=block_identifier)

    def add_active_raffle_type(self, cred:Credentials, _id:uint256, _bucket:uint256) -> TxReceipt:
        tx = self.contract.functions.addActiveRaffleType(_id, _bucket)
        return self.send_transaction(tx, cred)

    def add_and_activate_raffle_type(self, cred:Credentials, _rewards:Sequence[address], _reward_amounts:Sequence[uint256], _max_winners:uint256, _duration:uint64, _buckets:Sequence[uint256]) -> TxReceipt:
        tx = self.contract.functions.addAndActivateRaffleType(_rewards, _reward_amounts, _max_winners, _duration, _buckets)
        return self.send_transaction(tx, cred)

    def add_raffle_type(self, cred:Credentials, _rewards:Sequence[address], _reward_amounts:Sequence[uint256], _max_winners:uint256, _duration:uint64) -> TxReceipt:
        tx = self.contract.functions.addRaffleType(_rewards, _reward_amounts, _max_winners, _duration)
        return self.send_transaction(tx, cred)

    def close_raffles(self, cred:Credentials) -> TxReceipt:
        tx = self.contract.functions.closeRaffles()
        return self.send_transaction(tx, cred)

    def create_raffle(self, cred:Credentials, _raffle_type_id:uint256, _bucket:uint256) -> TxReceipt:
        tx = self.contract.functions.createRaffle(_raffle_type_id, _bucket)
        return self.send_transaction(tx, cred)

    def current_raffle_buckets(self, a:uint256, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.currentRaffleBuckets(a).call(block_identifier=block_identifier)

    def draw_winners(self, cred:Credentials) -> TxReceipt:
        tx = self.contract.functions.drawWinners()
        return self.send_transaction(tx, cred)

    def enter_raffle(self, cred:Credentials, _raffle_id:uint256, _tickets:uint256) -> TxReceipt:
        tx = self.contract.functions.enterRaffle(_raffle_id, _tickets)
        return self.send_transaction(tx, cred)

    def entries(self, a:uint256, b:uint256, block_identifier:BlockIdentifier = 'latest') -> address:
        return self.contract.functions.entries(a, b).call(block_identifier=block_identifier)

    def get_current_raffle_buckets(self, block_identifier:BlockIdentifier = 'latest') -> List[uint256]:
        return self.contract.functions.getCurrentRaffleBuckets().call(block_identifier=block_identifier)

    def get_current_raffle_data(self, block_identifier:BlockIdentifier = 'latest') -> Tuple[List[tuple], List[tuple], List[uint256], List[uint256]]:
        return self.contract.functions.getCurrentRaffleData().call(block_identifier=block_identifier)

    def get_last_raffle_buckets(self, block_identifier:BlockIdentifier = 'latest') -> List[uint256]:
        return self.contract.functions.getLastRaffleBuckets().call(block_identifier=block_identifier)

    def get_previous_raffle_data(self, block_identifier:BlockIdentifier = 'latest') -> Tuple[List[tuple], List[tuple], List[uint256], List[uint256]]:
        return self.contract.functions.getPreviousRaffleData().call(block_identifier=block_identifier)

    def get_raffle_list(self, _raffle_buckets:Sequence[uint256], block_identifier:BlockIdentifier = 'latest') -> List[tuple]:
        return self.contract.functions.getRaffleList(_raffle_buckets).call(block_identifier=block_identifier)

    def get_raffle_tickets_allowance_list(self, _raffle_buckets:Sequence[uint256], block_identifier:BlockIdentifier = 'latest') -> List[uint256]:
        return self.contract.functions.getRaffleTicketsAllowanceList(_raffle_buckets).call(block_identifier=block_identifier)

    def get_raffle_tickets_list(self, _raffle_buckets:Sequence[uint256], block_identifier:BlockIdentifier = 'latest') -> List[uint256]:
        return self.contract.functions.getRaffleTicketsList(_raffle_buckets).call(block_identifier=block_identifier)

    def get_raffle_types_list(self, _raffle_buckets:Sequence[uint256], block_identifier:BlockIdentifier = 'latest') -> List[tuple]:
        return self.contract.functions.getRaffleTypesList(_raffle_buckets).call(block_identifier=block_identifier)

    def get_ticket_allowance(self, _user:address, _raffle_id:uint256, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.getTicketAllowance(_user, _raffle_id).call(block_identifier=block_identifier)

    def init_raffle_buckets(self, cred:Credentials, _buckets:uint256) -> TxReceipt:
        tx = self.contract.functions.initRaffleBuckets(_buckets)
        return self.send_transaction(tx, cred)

    def initialize(self, cred:Credentials, _raffle_tickets:address, _airdrop_claim:address, _item_minter:address) -> TxReceipt:
        tx = self.contract.functions.initialize(_raffle_tickets, _airdrop_claim, _item_minter)
        return self.send_transaction(tx, cred)

    def last_raffle_buckets(self, a:uint256, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.lastRaffleBuckets(a).call(block_identifier=block_identifier)

    def pause(self, cred:Credentials) -> TxReceipt:
        tx = self.contract.functions.pause()
        return self.send_transaction(tx, cred)

    def paused(self, block_identifier:BlockIdentifier = 'latest') -> bool:
        return self.contract.functions.paused().call(block_identifier=block_identifier)

    def player_entries(self, a:uint256, b:address, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.playerEntries(a, b).call(block_identifier=block_identifier)

    def raffle_types(self, a:uint256, block_identifier:BlockIdentifier = 'latest') -> Tuple[uint256, uint256, uint64]:
        return self.contract.functions.raffleTypes(a).call(block_identifier=block_identifier)

    def raffles(self, a:uint256, block_identifier:BlockIdentifier = 'latest') -> Tuple[uint256, uint256, uint256, uint256, uint256, uint256, uint8]:
        return self.contract.functions.raffles(a).call(block_identifier=block_identifier)

    def remove_active_raffle_type(self, cred:Credentials, _index:uint256, _bucket:uint256) -> TxReceipt:
        tx = self.contract.functions.removeActiveRaffleType(_index, _bucket)
        return self.send_transaction(tx, cred)

    def set_raffle_type(self, cred:Credentials, _raffle_type_id:uint256, _rewards:Sequence[address], _reward_amounts:Sequence[uint256], _max_winners:uint256, _duration:uint64) -> TxReceipt:
        tx = self.contract.functions.setRaffleType(_raffle_type_id, _rewards, _reward_amounts, _max_winners, _duration)
        return self.send_transaction(tx, cred)

    def start_random_raffle(self, cred:Credentials, _bucket:uint256) -> TxReceipt:
        tx = self.contract.functions.startRandomRaffle(_bucket)
        return self.send_transaction(tx, cred)

    def start_random_raffles(self, cred:Credentials, _buckets:Sequence[uint256]) -> TxReceipt:
        tx = self.contract.functions.startRandomRaffles(_buckets)
        return self.send_transaction(tx, cred)

    def time_per_ticket(self, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.timePerTicket().call(block_identifier=block_identifier)

    def total_raffle_types(self, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.totalRaffleTypes().call(block_identifier=block_identifier)

    def total_raffles(self, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.totalRaffles().call(block_identifier=block_identifier)

    def unpause(self, cred:Credentials) -> TxReceipt:
        tx = self.contract.functions.unpause()
        return self.send_transaction(tx, cred)



from ..abi_contract_wrapper import ABIContractWrapper
from ..solidity_types import *
from ..credentials import Credentials

CONTRACT_ADDRESS =     {
    "cv": "0x739B1666c2956f601f095298132773074c3E184b",
    "sd": "0xEE258eF5F4338B37E9BA9dE6a56382AdB32056E2"
}

ABI = """[
    {"name": "AdminChanged", "type": "event", "inputs": [{"name": "previousAdmin", "type": "address", "indexed": false, "internalType": "address"}, {"name": "newAdmin", "type": "address", "indexed": false, "internalType": "address"}], "anonymous": false},
    {"name": "Upgraded", "type": "event", "inputs": [{"name": "implementation", "type": "address", "indexed": true, "internalType": "address"}], "anonymous": false},
    {"type": "fallback", "stateMutability": "payable"},
    {"name": "admin", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "address", "internalType": "address"}], "stateMutability": "nonpayable"},
    {"name": "changeAdmin", "type": "function", "inputs": [{"name": "newAdmin", "type": "address", "internalType": "address"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "implementation", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "address", "internalType": "address"}], "stateMutability": "nonpayable"},
    {"name": "upgradeTo", "type": "function", "inputs": [{"name": "newImplementation", "type": "address", "internalType": "address"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "upgradeToAndCall", "type": "function", "inputs": [{"name": "newImplementation", "type": "address", "internalType": "address"}, {"name": "data", "type": "bytes", "internalType": "bytes"}], "outputs": [], "stateMutability": "payable"},
    {"type": "receive", "stateMutability": "payable"},
    {"name": "HeroArrived", "type": "event", "inputs": [{"name": "heroId", "type": "uint256", "indexed": true, "internalType": "uint256"}, {"name": "arrivalChainId", "type": "uint256", "indexed": false, "internalType": "uint256"}], "anonymous": false},
    {"name": "HeroSent", "type": "event", "inputs": [{"name": "heroId", "type": "uint256", "indexed": true, "internalType": "uint256"}, {"name": "arrivalChainId", "type": "uint256", "indexed": false, "internalType": "uint256"}], "anonymous": false},
    {"name": "OwnershipTransferred", "type": "event", "inputs": [{"name": "previousOwner", "type": "address", "indexed": true, "internalType": "address"}, {"name": "newOwner", "type": "address", "indexed": true, "internalType": "address"}], "anonymous": false},
    {"name": "SetTrustedRemote", "type": "event", "inputs": [{"name": "_srcChainId", "type": "uint256", "indexed": false, "internalType": "uint256"}, {"name": "_srcAddress", "type": "bytes32", "indexed": false, "internalType": "bytes32"}], "anonymous": false},
    {"name": "assistingAuction", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "address", "internalType": "address"}], "stateMutability": "view"},
    {"name": "executeMessage", "type": "function", "inputs": [{"name": "_srcAddress", "type": "bytes32", "internalType": "bytes32"}, {"name": "_srcChainId", "type": "uint256", "internalType": "uint256"}, {"name": "_message", "type": "bytes", "internalType": "bytes"}, {"name": "_executor", "type": "address", "internalType": "address"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "getTrustedRemote", "type": "function", "inputs": [{"name": "_chainId", "type": "uint256", "internalType": "uint256"}], "outputs": [{"name": "trustedRemote", "type": "bytes32", "internalType": "bytes32"}], "stateMutability": "view"},
    {"name": "heroes", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "address", "internalType": "address"}], "stateMutability": "view"},
    {"name": "initialize", "type": "function", "inputs": [{"name": "_messageBus", "type": "address", "internalType": "address"}, {"name": "_heroes", "type": "address", "internalType": "address"}, {"name": "_assistingAuction", "type": "address", "internalType": "address"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "messageBus", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "address", "internalType": "address"}], "stateMutability": "view"},
    {"name": "msgGasLimit", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "owner", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "address", "internalType": "address"}], "stateMutability": "view"},
    {"name": "renounceOwnership", "type": "function", "inputs": [], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "sendHero", "type": "function", "inputs": [{"name": "_heroId", "type": "uint256", "internalType": "uint256"}, {"name": "_dstChainId", "type": "uint256", "internalType": "uint256"}], "outputs": [], "stateMutability": "payable"},
    {"name": "setAssistingAuctionAddress", "type": "function", "inputs": [{"name": "_assistingAuction", "type": "address", "internalType": "address"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "setMessageBus", "type": "function", "inputs": [{"name": "_messageBus", "type": "address", "internalType": "address"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "setMsgGasLimit", "type": "function", "inputs": [{"name": "_msgGasLimit", "type": "uint256", "internalType": "uint256"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "setTrustedRemote", "type": "function", "inputs": [{"name": "_srcChainId", "type": "uint256", "internalType": "uint256"}, {"name": "_srcAddress", "type": "bytes32", "internalType": "bytes32"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "transferOwnership", "type": "function", "inputs": [{"name": "newOwner", "type": "address", "internalType": "address"}], "outputs": [], "stateMutability": "nonpayable"},
    {"type": "constructor", "inputs": [{"name": "initialLogic", "type": "address", "internalType": "address"}, {"name": "initialAdmin", "type": "address", "internalType": "address"}, {"name": "_data", "type": "bytes", "internalType": "bytes"}], "stateMutability": "payable"}
]
"""     

class HeroBridge(ABIContractWrapper):
    def __init__(self, chain_key:str, rpc:str):
        contract_address = CONTRACT_ADDRESS[chain_key]
        super().__init__(contract_address=contract_address, abi=ABI, rpc=rpc)

    def admin(self, cred:Credentials) -> TxReceipt:
        tx = self.contract.functions.admin()
        return self.send_transaction(tx, cred)

    def change_admin(self, cred:Credentials, new_admin:address) -> TxReceipt:
        tx = self.contract.functions.changeAdmin(new_admin)
        return self.send_transaction(tx, cred)

    def implementation(self, cred:Credentials) -> TxReceipt:
        tx = self.contract.functions.implementation()
        return self.send_transaction(tx, cred)

    def upgrade_to(self, cred:Credentials, new_implementation:address) -> TxReceipt:
        tx = self.contract.functions.upgradeTo(new_implementation)
        return self.send_transaction(tx, cred)

    def upgrade_to_and_call(self, cred:Credentials, new_implementation:address, data:bytes) -> TxReceipt:
        tx = self.contract.functions.upgradeToAndCall(new_implementation, data)
        return self.send_transaction(tx, cred)

    def assisting_auction(self, block_identifier:BlockIdentifier = 'latest') -> address:
        return self.contract.functions.assistingAuction().call(block_identifier=block_identifier)

    def execute_message(self, cred:Credentials, _src_address:bytes32, _src_chain_id:uint256, _message:bytes, _executor:address) -> TxReceipt:
        tx = self.contract.functions.executeMessage(_src_address, _src_chain_id, _message, _executor)
        return self.send_transaction(tx, cred)

    def get_trusted_remote(self, _chain_id:uint256, block_identifier:BlockIdentifier = 'latest') -> bytes32:
        return self.contract.functions.getTrustedRemote(_chain_id).call(block_identifier=block_identifier)

    def heroes(self, block_identifier:BlockIdentifier = 'latest') -> address:
        return self.contract.functions.heroes().call(block_identifier=block_identifier)

    def initialize(self, cred:Credentials, _message_bus:address, _heroes:address, _assisting_auction:address) -> TxReceipt:
        tx = self.contract.functions.initialize(_message_bus, _heroes, _assisting_auction)
        return self.send_transaction(tx, cred)

    def message_bus(self, block_identifier:BlockIdentifier = 'latest') -> address:
        return self.contract.functions.messageBus().call(block_identifier=block_identifier)

    def msg_gas_limit(self, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.msgGasLimit().call(block_identifier=block_identifier)

    def owner(self, block_identifier:BlockIdentifier = 'latest') -> address:
        return self.contract.functions.owner().call(block_identifier=block_identifier)

    def renounce_ownership(self, cred:Credentials) -> TxReceipt:
        tx = self.contract.functions.renounceOwnership()
        return self.send_transaction(tx, cred)

    def send_hero(self, cred:Credentials, _hero_id:uint256, _dst_chain_id:uint256) -> TxReceipt:
        tx = self.contract.functions.sendHero(_hero_id, _dst_chain_id)
        return self.send_transaction(tx, cred)

    def set_assisting_auction_address(self, cred:Credentials, _assisting_auction:address) -> TxReceipt:
        tx = self.contract.functions.setAssistingAuctionAddress(_assisting_auction)
        return self.send_transaction(tx, cred)

    def set_message_bus(self, cred:Credentials, _message_bus:address) -> TxReceipt:
        tx = self.contract.functions.setMessageBus(_message_bus)
        return self.send_transaction(tx, cred)

    def set_msg_gas_limit(self, cred:Credentials, _msg_gas_limit:uint256) -> TxReceipt:
        tx = self.contract.functions.setMsgGasLimit(_msg_gas_limit)
        return self.send_transaction(tx, cred)

    def set_trusted_remote(self, cred:Credentials, _src_chain_id:uint256, _src_address:bytes32) -> TxReceipt:
        tx = self.contract.functions.setTrustedRemote(_src_chain_id, _src_address)
        return self.send_transaction(tx, cred)

    def transfer_ownership(self, cred:Credentials, new_owner:address) -> TxReceipt:
        tx = self.contract.functions.transferOwnership(new_owner)
        return self.send_transaction(tx, cred)



from ..abi_contract_wrapper import ABIContractWrapper
from ..solidity_types import *
from ..credentials import Credentials

CONTRACT_ADDRESS =     {
    "cv": "0xc9A9F352Aa188f422A8f8902B547FB3E59D37210",
    "sd": "0xF78cA21d7Da3227457138714F5bEd08D2604A156"
}

ABI = """[
    {"name": "Initialized", "type": "event", "inputs": [{"name": "version", "type": "uint8", "indexed": false, "internalType": "uint8"}], "anonymous": false},
    {"name": "ItemConsumed", "type": "event", "inputs": [{"name": "player", "type": "address", "indexed": false, "internalType": "address"}, {"name": "item", "type": "address", "indexed": false, "internalType": "address"}, {"name": "heroId", "type": "uint256", "indexed": false, "internalType": "uint256"}, {"name": "oldHero", "type": "tuple", "components": [{"name": "id", "type": "uint256", "internalType": "uint256"}, {"name": "summoningInfo", "type": "tuple", "components": [{"name": "summonedTime", "type": "uint256", "internalType": "uint256"}, {"name": "nextSummonTime", "type": "uint256", "internalType": "uint256"}, {"name": "summonerId", "type": "uint256", "internalType": "uint256"}, {"name": "assistantId", "type": "uint256", "internalType": "uint256"}, {"name": "summons", "type": "uint32", "internalType": "uint32"}, {"name": "maxSummons", "type": "uint32", "internalType": "uint32"}], "internalType": "struct IHeroTypes.SummoningInfo"}, {"name": "info", "type": "tuple", "components": [{"name": "statGenes", "type": "uint256", "internalType": "uint256"}, {"name": "visualGenes", "type": "uint256", "internalType": "uint256"}, {"name": "rarity", "type": "uint8", "internalType": "enum IHeroTypes.Rarity"}, {"name": "shiny", "type": "bool", "internalType": "bool"}, {"name": "generation", "type": "uint16", "internalType": "uint16"}, {"name": "firstName", "type": "uint32", "internalType": "uint32"}, {"name": "lastName", "type": "uint32", "internalType": "uint32"}, {"name": "shinyStyle", "type": "uint8", "internalType": "uint8"}, {"name": "class", "type": "uint8", "internalType": "uint8"}, {"name": "subClass", "type": "uint8", "internalType": "uint8"}], "internalType": "struct IHeroTypes.HeroInfo"}, {"name": "state", "type": "tuple", "components": [{"name": "staminaFullAt", "type": "uint256", "internalType": "uint256"}, {"name": "hpFullAt", "type": "uint256", "internalType": "uint256"}, {"name": "mpFullAt", "type": "uint256", "internalType": "uint256"}, {"name": "level", "type": "uint16", "internalType": "uint16"}, {"name": "xp", "type": "uint64", "internalType": "uint64"}, {"name": "currentQuest", "type": "address", "internalType": "address"}, {"name": "sp", "type": "uint8", "internalType": "uint8"}, {"name": "status", "type": "uint8", "internalType": "enum IHeroTypes.HeroStatus"}], "internalType": "struct IHeroTypes.HeroState"}, {"name": "stats", "type": "tuple", "components": [{"name": "strength", "type": "uint16", "internalType": "uint16"}, {"name": "intelligence", "type": "uint16", "internalType": "uint16"}, {"name": "wisdom", "type": "uint16", "internalType": "uint16"}, {"name": "luck", "type": "uint16", "internalType": "uint16"}, {"name": "agility", "type": "uint16", "internalType": "uint16"}, {"name": "vitality", "type": "uint16", "internalType": "uint16"}, {"name": "endurance", "type": "uint16", "internalType": "uint16"}, {"name": "dexterity", "type": "uint16", "internalType": "uint16"}, {"name": "hp", "type": "uint16", "internalType": "uint16"}, {"name": "mp", "type": "uint16", "internalType": "uint16"}, {"name": "stamina", "type": "uint16", "internalType": "uint16"}], "internalType": "struct IHeroTypes.HeroStats"}, {"name": "primaryStatGrowth", "type": "tuple", "components": [{"name": "strength", "type": "uint16", "internalType": "uint16"}, {"name": "intelligence", "type": "uint16", "internalType": "uint16"}, {"name": "wisdom", "type": "uint16", "internalType": "uint16"}, {"name": "luck", "type": "uint16", "internalType": "uint16"}, {"name": "agility", "type": "uint16", "internalType": "uint16"}, {"name": "vitality", "type": "uint16", "internalType": "uint16"}, {"name": "endurance", "type": "uint16", "internalType": "uint16"}, {"name": "dexterity", "type": "uint16", "internalType": "uint16"}, {"name": "hpSm", "type": "uint16", "internalType": "uint16"}, {"name": "hpRg", "type": "uint16", "internalType": "uint16"}, {"name": "hpLg", "type": "uint16", "internalType": "uint16"}, {"name": "mpSm", "type": "uint16", "internalType": "uint16"}, {"name": "mpRg", "type": "uint16", "internalType": "uint16"}, {"name": "mpLg", "type": "uint16", "internalType": "uint16"}], "internalType": "struct IHeroTypes.HeroStatGrowth"}, {"name": "secondaryStatGrowth", "type": "tuple", "components": [{"name": "strength", "type": "uint16", "internalType": "uint16"}, {"name": "intelligence", "type": "uint16", "internalType": "uint16"}, {"name": "wisdom", "type": "uint16", "internalType": "uint16"}, {"name": "luck", "type": "uint16", "internalType": "uint16"}, {"name": "agility", "type": "uint16", "internalType": "uint16"}, {"name": "vitality", "type": "uint16", "internalType": "uint16"}, {"name": "endurance", "type": "uint16", "internalType": "uint16"}, {"name": "dexterity", "type": "uint16", "internalType": "uint16"}, {"name": "hpSm", "type": "uint16", "internalType": "uint16"}, {"name": "hpRg", "type": "uint16", "internalType": "uint16"}, {"name": "hpLg", "type": "uint16", "internalType": "uint16"}, {"name": "mpSm", "type": "uint16", "internalType": "uint16"}, {"name": "mpRg", "type": "uint16", "internalType": "uint16"}, {"name": "mpLg", "type": "uint16", "internalType": "uint16"}], "internalType": "struct IHeroTypes.HeroStatGrowth"}, {"name": "professions", "type": "tuple", "components": [{"name": "mining", "type": "uint16", "internalType": "uint16"}, {"name": "gardening", "type": "uint16", "internalType": "uint16"}, {"name": "foraging", "type": "uint16", "internalType": "uint16"}, {"name": "fishing", "type": "uint16", "internalType": "uint16"}], "internalType": "struct IHeroTypes.HeroProfessions"}], "indexed": false, "internalType": "struct IHeroTypes.Hero"}, {"name": "newHero", "type": "tuple", "components": [{"name": "id", "type": "uint256", "internalType": "uint256"}, {"name": "summoningInfo", "type": "tuple", "components": [{"name": "summonedTime", "type": "uint256", "internalType": "uint256"}, {"name": "nextSummonTime", "type": "uint256", "internalType": "uint256"}, {"name": "summonerId", "type": "uint256", "internalType": "uint256"}, {"name": "assistantId", "type": "uint256", "internalType": "uint256"}, {"name": "summons", "type": "uint32", "internalType": "uint32"}, {"name": "maxSummons", "type": "uint32", "internalType": "uint32"}], "internalType": "struct IHeroTypes.SummoningInfo"}, {"name": "info", "type": "tuple", "components": [{"name": "statGenes", "type": "uint256", "internalType": "uint256"}, {"name": "visualGenes", "type": "uint256", "internalType": "uint256"}, {"name": "rarity", "type": "uint8", "internalType": "enum IHeroTypes.Rarity"}, {"name": "shiny", "type": "bool", "internalType": "bool"}, {"name": "generation", "type": "uint16", "internalType": "uint16"}, {"name": "firstName", "type": "uint32", "internalType": "uint32"}, {"name": "lastName", "type": "uint32", "internalType": "uint32"}, {"name": "shinyStyle", "type": "uint8", "internalType": "uint8"}, {"name": "class", "type": "uint8", "internalType": "uint8"}, {"name": "subClass", "type": "uint8", "internalType": "uint8"}], "internalType": "struct IHeroTypes.HeroInfo"}, {"name": "state", "type": "tuple", "components": [{"name": "staminaFullAt", "type": "uint256", "internalType": "uint256"}, {"name": "hpFullAt", "type": "uint256", "internalType": "uint256"}, {"name": "mpFullAt", "type": "uint256", "internalType": "uint256"}, {"name": "level", "type": "uint16", "internalType": "uint16"}, {"name": "xp", "type": "uint64", "internalType": "uint64"}, {"name": "currentQuest", "type": "address", "internalType": "address"}, {"name": "sp", "type": "uint8", "internalType": "uint8"}, {"name": "status", "type": "uint8", "internalType": "enum IHeroTypes.HeroStatus"}], "internalType": "struct IHeroTypes.HeroState"}, {"name": "stats", "type": "tuple", "components": [{"name": "strength", "type": "uint16", "internalType": "uint16"}, {"name": "intelligence", "type": "uint16", "internalType": "uint16"}, {"name": "wisdom", "type": "uint16", "internalType": "uint16"}, {"name": "luck", "type": "uint16", "internalType": "uint16"}, {"name": "agility", "type": "uint16", "internalType": "uint16"}, {"name": "vitality", "type": "uint16", "internalType": "uint16"}, {"name": "endurance", "type": "uint16", "internalType": "uint16"}, {"name": "dexterity", "type": "uint16", "internalType": "uint16"}, {"name": "hp", "type": "uint16", "internalType": "uint16"}, {"name": "mp", "type": "uint16", "internalType": "uint16"}, {"name": "stamina", "type": "uint16", "internalType": "uint16"}], "internalType": "struct IHeroTypes.HeroStats"}, {"name": "primaryStatGrowth", "type": "tuple", "components": [{"name": "strength", "type": "uint16", "internalType": "uint16"}, {"name": "intelligence", "type": "uint16", "internalType": "uint16"}, {"name": "wisdom", "type": "uint16", "internalType": "uint16"}, {"name": "luck", "type": "uint16", "internalType": "uint16"}, {"name": "agility", "type": "uint16", "internalType": "uint16"}, {"name": "vitality", "type": "uint16", "internalType": "uint16"}, {"name": "endurance", "type": "uint16", "internalType": "uint16"}, {"name": "dexterity", "type": "uint16", "internalType": "uint16"}, {"name": "hpSm", "type": "uint16", "internalType": "uint16"}, {"name": "hpRg", "type": "uint16", "internalType": "uint16"}, {"name": "hpLg", "type": "uint16", "internalType": "uint16"}, {"name": "mpSm", "type": "uint16", "internalType": "uint16"}, {"name": "mpRg", "type": "uint16", "internalType": "uint16"}, {"name": "mpLg", "type": "uint16", "internalType": "uint16"}], "internalType": "struct IHeroTypes.HeroStatGrowth"}, {"name": "secondaryStatGrowth", "type": "tuple", "components": [{"name": "strength", "type": "uint16", "internalType": "uint16"}, {"name": "intelligence", "type": "uint16", "internalType": "uint16"}, {"name": "wisdom", "type": "uint16", "internalType": "uint16"}, {"name": "luck", "type": "uint16", "internalType": "uint16"}, {"name": "agility", "type": "uint16", "internalType": "uint16"}, {"name": "vitality", "type": "uint16", "internalType": "uint16"}, {"name": "endurance", "type": "uint16", "internalType": "uint16"}, {"name": "dexterity", "type": "uint16", "internalType": "uint16"}, {"name": "hpSm", "type": "uint16", "internalType": "uint16"}, {"name": "hpRg", "type": "uint16", "internalType": "uint16"}, {"name": "hpLg", "type": "uint16", "internalType": "uint16"}, {"name": "mpSm", "type": "uint16", "internalType": "uint16"}, {"name": "mpRg", "type": "uint16", "internalType": "uint16"}, {"name": "mpLg", "type": "uint16", "internalType": "uint16"}], "internalType": "struct IHeroTypes.HeroStatGrowth"}, {"name": "professions", "type": "tuple", "components": [{"name": "mining", "type": "uint16", "internalType": "uint16"}, {"name": "gardening", "type": "uint16", "internalType": "uint16"}, {"name": "foraging", "type": "uint16", "internalType": "uint16"}, {"name": "fishing", "type": "uint16", "internalType": "uint16"}], "internalType": "struct IHeroTypes.HeroProfessions"}], "indexed": false, "internalType": "struct IHeroTypes.Hero"}], "anonymous": false},
    {"name": "consumeItem", "type": "function", "inputs": [{"name": "_consumableAddress", "type": "address", "internalType": "address"}, {"name": "_heroId", "type": "uint256", "internalType": "uint256"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "initialize", "type": "function", "inputs": [{"name": "_heroCoreAddress", "type": "address", "internalType": "address"}], "outputs": [], "stateMutability": "nonpayable"}
]
"""     

class ItemConsumer(ABIContractWrapper):
    def __init__(self, chain_key:str, rpc:str):
        contract_address = CONTRACT_ADDRESS[chain_key]
        super().__init__(contract_address=contract_address, abi=ABI, rpc=rpc)

    def consume_item(self, cred:Credentials, _consumable_address:address, _hero_id:uint256) -> TxReceipt:
        tx = self.contract.functions.consumeItem(_consumable_address, _hero_id)
        return self.send_transaction(tx, cred)

    def initialize(self, cred:Credentials, _hero_core_address:address) -> TxReceipt:
        tx = self.contract.functions.initialize(_hero_core_address)
        return self.send_transaction(tx, cred)



from ..abi_contract_wrapper import ABIContractWrapper
from ..solidity_types import *
from ..credentials import Credentials

CONTRACT_ADDRESS =     {
    "cv": "0xABABB0A2c42274D0e81417B824CABca464F5c16C",
    "sd": "0x0000000000000000000000000000000000000000"
}

ABI = """[
    {"name": "Donation", "type": "event", "inputs": [{"name": "account", "type": "address", "indexed": true, "internalType": "address"}, {"name": "token", "type": "address", "indexed": true, "internalType": "address"}, {"name": "amount", "type": "uint256", "indexed": false, "internalType": "uint256"}], "anonymous": false},
    {"name": "Initialized", "type": "event", "inputs": [{"name": "version", "type": "uint8", "indexed": false, "internalType": "uint8"}], "anonymous": false},
    {"name": "OwnershipTransferred", "type": "event", "inputs": [{"name": "previousOwner", "type": "address", "indexed": true, "internalType": "address"}, {"name": "newOwner", "type": "address", "indexed": true, "internalType": "address"}], "anonymous": false},
    {"name": "Withdrawal", "type": "event", "inputs": [{"name": "token", "type": "address", "indexed": true, "internalType": "address"}, {"name": "amount", "type": "uint256", "indexed": false, "internalType": "uint256"}], "anonymous": false},
    {"type": "fallback", "stateMutability": "payable"},
    {"name": "accepted", "type": "function", "inputs": [{"name": "", "type": "address", "internalType": "address"}], "outputs": [{"name": "", "type": "bool", "internalType": "bool"}], "stateMutability": "view"},
    {"name": "addToken", "type": "function", "inputs": [{"name": "_token", "type": "address", "internalType": "address"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "campaignId", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "uint8", "internalType": "uint8"}], "stateMutability": "view"},
    {"name": "disableToken", "type": "function", "inputs": [{"name": "_token", "type": "address", "internalType": "address"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "donateNativeToken", "type": "function", "inputs": [], "outputs": [], "stateMutability": "payable"},
    {"name": "donateToken", "type": "function", "inputs": [{"name": "_token", "type": "address", "internalType": "address"}, {"name": "_wei", "type": "uint256", "internalType": "uint256"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "donationTotals", "type": "function", "inputs": [{"name": "", "type": "address", "internalType": "address"}], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "enableToken", "type": "function", "inputs": [{"name": "_token", "type": "address", "internalType": "address"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "getInfo", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "tuple[]", "components": [{"name": "accepted", "type": "bool", "internalType": "bool"}, {"name": "token", "type": "address", "internalType": "address"}, {"name": "name", "type": "string", "internalType": "string"}, {"name": "symbol", "type": "string", "internalType": "string"}, {"name": "decimals", "type": "uint256", "internalType": "uint256"}, {"name": "total", "type": "uint256", "internalType": "uint256"}, {"name": "balance", "type": "uint256", "internalType": "uint256"}, {"name": "userDonation", "type": "uint256", "internalType": "uint256"}, {"name": "campaignDonationHistory", "type": "uint256", "internalType": "uint256"}], "internalType": "struct Structs.Info[]"}], "stateMutability": "view"},
    {"name": "initialize", "type": "function", "inputs": [{"name": "_multisig", "type": "address", "internalType": "address"}, {"name": "_flagStorageAddress", "type": "address", "internalType": "address"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "multisig", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "address", "internalType": "address"}], "stateMutability": "view"},
    {"name": "owner", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "address", "internalType": "address"}], "stateMutability": "view"},
    {"name": "renounceOwnership", "type": "function", "inputs": [], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "setCampaignId", "type": "function", "inputs": [{"name": "_campaignId", "type": "uint8", "internalType": "uint8"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "tokens", "type": "function", "inputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "outputs": [{"name": "", "type": "address", "internalType": "address"}], "stateMutability": "view"},
    {"name": "transferOwnership", "type": "function", "inputs": [{"name": "newOwner", "type": "address", "internalType": "address"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "withdraw", "type": "function", "inputs": [{"name": "_token", "type": "address", "internalType": "address"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "withdrawAll", "type": "function", "inputs": [], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "withdrawNative", "type": "function", "inputs": [], "outputs": [], "stateMutability": "nonpayable"},
    {"type": "receive", "stateMutability": "payable"}
]
"""     

class CharityFund(ABIContractWrapper):
    def __init__(self, chain_key:str, rpc:str):
        contract_address = CONTRACT_ADDRESS[chain_key]
        super().__init__(contract_address=contract_address, abi=ABI, rpc=rpc)

    def accepted(self, a:address, block_identifier:BlockIdentifier = 'latest') -> bool:
        return self.contract.functions.accepted(a).call(block_identifier=block_identifier)

    def add_token(self, cred:Credentials, _token:address) -> TxReceipt:
        tx = self.contract.functions.addToken(_token)
        return self.send_transaction(tx, cred)

    def campaign_id(self, block_identifier:BlockIdentifier = 'latest') -> uint8:
        return self.contract.functions.campaignId().call(block_identifier=block_identifier)

    def disable_token(self, cred:Credentials, _token:address) -> TxReceipt:
        tx = self.contract.functions.disableToken(_token)
        return self.send_transaction(tx, cred)

    def donate_native_token(self, cred:Credentials) -> TxReceipt:
        tx = self.contract.functions.donateNativeToken()
        return self.send_transaction(tx, cred)

    def donate_token(self, cred:Credentials, _token:address, _wei:uint256) -> TxReceipt:
        tx = self.contract.functions.donateToken(_token, _wei)
        return self.send_transaction(tx, cred)

    def donation_totals(self, a:address, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.donationTotals(a).call(block_identifier=block_identifier)

    def enable_token(self, cred:Credentials, _token:address) -> TxReceipt:
        tx = self.contract.functions.enableToken(_token)
        return self.send_transaction(tx, cred)

    def get_info(self, block_identifier:BlockIdentifier = 'latest') -> List[tuple]:
        return self.contract.functions.getInfo().call(block_identifier=block_identifier)

    def initialize(self, cred:Credentials, _multisig:address, _flag_storage_address:address) -> TxReceipt:
        tx = self.contract.functions.initialize(_multisig, _flag_storage_address)
        return self.send_transaction(tx, cred)

    def multisig(self, block_identifier:BlockIdentifier = 'latest') -> address:
        return self.contract.functions.multisig().call(block_identifier=block_identifier)

    def owner(self, block_identifier:BlockIdentifier = 'latest') -> address:
        return self.contract.functions.owner().call(block_identifier=block_identifier)

    def renounce_ownership(self, cred:Credentials) -> TxReceipt:
        tx = self.contract.functions.renounceOwnership()
        return self.send_transaction(tx, cred)

    def set_campaign_id(self, cred:Credentials, _campaign_id:uint8) -> TxReceipt:
        tx = self.contract.functions.setCampaignId(_campaign_id)
        return self.send_transaction(tx, cred)

    def tokens(self, a:uint256, block_identifier:BlockIdentifier = 'latest') -> address:
        return self.contract.functions.tokens(a).call(block_identifier=block_identifier)

    def transfer_ownership(self, cred:Credentials, new_owner:address) -> TxReceipt:
        tx = self.contract.functions.transferOwnership(new_owner)
        return self.send_transaction(tx, cred)

    def withdraw(self, cred:Credentials, _token:address) -> TxReceipt:
        tx = self.contract.functions.withdraw(_token)
        return self.send_transaction(tx, cred)

    def withdraw_all(self, cred:Credentials) -> TxReceipt:
        tx = self.contract.functions.withdrawAll()
        return self.send_transaction(tx, cred)

    def withdraw_native(self, cred:Credentials) -> TxReceipt:
        tx = self.contract.functions.withdrawNative()
        return self.send_transaction(tx, cred)



from ..abi_contract_wrapper import ABIContractWrapper
from ..solidity_types import *
from ..credentials import Credentials

CONTRACT_ADDRESS =     {
    "cv": "0x4b1F4C7981465F814c4A78d79be21558A60f57F2",
    "sd": "0x0000000000000000000000000000000000000000"
}

ABI = """[
    {"type": "constructor", "inputs": [{"name": "_factory", "type": "address", "internalType": "address"}, {"name": "_bank", "type": "address", "internalType": "address"}, {"name": "_govToken", "type": "address", "internalType": "address"}, {"name": "_weth", "type": "address", "internalType": "address"}], "stateMutability": "nonpayable"},
    {"name": "LogBridgeSet", "type": "event", "inputs": [{"name": "token", "type": "address", "indexed": true, "internalType": "address"}, {"name": "bridge", "type": "address", "indexed": true, "internalType": "address"}], "anonymous": false},
    {"name": "LogConvert", "type": "event", "inputs": [{"name": "server", "type": "address", "indexed": true, "internalType": "address"}, {"name": "token0", "type": "address", "indexed": true, "internalType": "address"}, {"name": "token1", "type": "address", "indexed": true, "internalType": "address"}, {"name": "amount0", "type": "uint256", "indexed": false, "internalType": "uint256"}, {"name": "amount1", "type": "uint256", "indexed": false, "internalType": "uint256"}, {"name": "amountGovToken", "type": "uint256", "indexed": false, "internalType": "uint256"}], "anonymous": false},
    {"name": "OwnershipTransferred", "type": "event", "inputs": [{"name": "previousOwner", "type": "address", "indexed": true, "internalType": "address"}, {"name": "newOwner", "type": "address", "indexed": true, "internalType": "address"}], "anonymous": false},
    {"name": "bank", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "address", "internalType": "address"}], "stateMutability": "view"},
    {"name": "bridgeFor", "type": "function", "inputs": [{"name": "token", "type": "address", "internalType": "address"}], "outputs": [{"name": "bridge", "type": "address", "internalType": "address"}], "stateMutability": "view"},
    {"name": "convert", "type": "function", "inputs": [{"name": "token0", "type": "address", "internalType": "address"}, {"name": "token1", "type": "address", "internalType": "address"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "convertMultiple", "type": "function", "inputs": [{"name": "token0", "type": "address[]", "internalType": "address[]"}, {"name": "token1", "type": "address[]", "internalType": "address[]"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "factory", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "address", "internalType": "contract IUniswapV2Factory"}], "stateMutability": "view"},
    {"name": "owner", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "address", "internalType": "address"}], "stateMutability": "view"},
    {"name": "renounceOwnership", "type": "function", "inputs": [], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "setBridge", "type": "function", "inputs": [{"name": "token", "type": "address", "internalType": "address"}, {"name": "bridge", "type": "address", "internalType": "address"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "transferOwnership", "type": "function", "inputs": [{"name": "newOwner", "type": "address", "internalType": "address"}], "outputs": [], "stateMutability": "nonpayable"}
]
"""     

class Banker(ABIContractWrapper):
    def __init__(self, chain_key:str, rpc:str):
        contract_address = CONTRACT_ADDRESS[chain_key]
        super().__init__(contract_address=contract_address, abi=ABI, rpc=rpc)

    def bank(self, block_identifier:BlockIdentifier = 'latest') -> address:
        return self.contract.functions.bank().call(block_identifier=block_identifier)

    def bridge_for(self, token:address, block_identifier:BlockIdentifier = 'latest') -> address:
        return self.contract.functions.bridgeFor(token).call(block_identifier=block_identifier)

    def convert(self, cred:Credentials, token0:address, token1:address) -> TxReceipt:
        tx = self.contract.functions.convert(token0, token1)
        return self.send_transaction(tx, cred)

    def convert_multiple(self, cred:Credentials, token0:Sequence[address], token1:Sequence[address]) -> TxReceipt:
        tx = self.contract.functions.convertMultiple(token0, token1)
        return self.send_transaction(tx, cred)

    def factory(self, block_identifier:BlockIdentifier = 'latest') -> address:
        return self.contract.functions.factory().call(block_identifier=block_identifier)

    def owner(self, block_identifier:BlockIdentifier = 'latest') -> address:
        return self.contract.functions.owner().call(block_identifier=block_identifier)

    def renounce_ownership(self, cred:Credentials) -> TxReceipt:
        tx = self.contract.functions.renounceOwnership()
        return self.send_transaction(tx, cred)

    def set_bridge(self, cred:Credentials, token:address, bridge:address) -> TxReceipt:
        tx = self.contract.functions.setBridge(token, bridge)
        return self.send_transaction(tx, cred)

    def transfer_ownership(self, cred:Credentials, new_owner:address) -> TxReceipt:
        tx = self.contract.functions.transferOwnership(new_owner)
        return self.send_transaction(tx, cred)



from ..abi_contract_wrapper import ABIContractWrapper
from ..solidity_types import *
from ..credentials import Credentials

CONTRACT_ADDRESS =     {
    "cv": "0x8101CfFBec8E045c3FAdC3877a1D30f97d301209",
    "sd": "0xA2cef1763e59198025259d76Ce8F9E60d27B17B5"
}

ABI = """[
    {"type": "constructor", "inputs": [{"name": "_heroCoreAddress", "type": "address", "internalType": "address"}, {"name": "_jewelAddress", "type": "address", "internalType": "address"}, {"name": "_cut", "type": "uint256", "internalType": "uint256"}], "stateMutability": "nonpayable"},
    {"name": "AuctionCancelled", "type": "event", "inputs": [{"name": "auctionId", "type": "uint256", "indexed": false, "internalType": "uint256"}, {"name": "tokenId", "type": "uint256", "indexed": true, "internalType": "uint256"}], "anonymous": false},
    {"name": "AuctionCreated", "type": "event", "inputs": [{"name": "auctionId", "type": "uint256", "indexed": false, "internalType": "uint256"}, {"name": "owner", "type": "address", "indexed": true, "internalType": "address"}, {"name": "tokenId", "type": "uint256", "indexed": true, "internalType": "uint256"}, {"name": "startingPrice", "type": "uint256", "indexed": false, "internalType": "uint256"}, {"name": "endingPrice", "type": "uint256", "indexed": false, "internalType": "uint256"}, {"name": "duration", "type": "uint256", "indexed": false, "internalType": "uint256"}, {"name": "winner", "type": "address", "indexed": false, "internalType": "address"}], "anonymous": false},
    {"name": "AuctionSuccessful", "type": "event", "inputs": [{"name": "auctionId", "type": "uint256", "indexed": false, "internalType": "uint256"}, {"name": "tokenId", "type": "uint256", "indexed": true, "internalType": "uint256"}, {"name": "totalPrice", "type": "uint256", "indexed": false, "internalType": "uint256"}, {"name": "winner", "type": "address", "indexed": false, "internalType": "address"}], "anonymous": false},
    {"name": "OwnershipTransferred", "type": "event", "inputs": [{"name": "previousOwner", "type": "address", "indexed": true, "internalType": "address"}, {"name": "newOwner", "type": "address", "indexed": true, "internalType": "address"}], "anonymous": false},
    {"name": "Paused", "type": "event", "inputs": [{"name": "account", "type": "address", "indexed": false, "internalType": "address"}], "anonymous": false},
    {"name": "Unpaused", "type": "event", "inputs": [{"name": "account", "type": "address", "indexed": false, "internalType": "address"}], "anonymous": false},
    {"name": "auctionHeroCore", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "address", "internalType": "contract IHeroCore"}], "stateMutability": "view"},
    {"name": "bid", "type": "function", "inputs": [{"name": "_tokenId", "type": "uint256", "internalType": "uint256"}, {"name": "_bidAmount", "type": "uint256", "internalType": "uint256"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "cancelAuction", "type": "function", "inputs": [{"name": "_tokenId", "type": "uint256", "internalType": "uint256"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "cancelAuctionWhenPaused", "type": "function", "inputs": [{"name": "_tokenId", "type": "uint256", "internalType": "uint256"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "createAuction", "type": "function", "inputs": [{"name": "_tokenId", "type": "uint256", "internalType": "uint256"}, {"name": "_startingPrice", "type": "uint128", "internalType": "uint128"}, {"name": "_endingPrice", "type": "uint128", "internalType": "uint128"}, {"name": "_duration", "type": "uint64", "internalType": "uint64"}, {"name": "_winner", "type": "address", "internalType": "address"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "getAuction", "type": "function", "inputs": [{"name": "_tokenId", "type": "uint256", "internalType": "uint256"}], "outputs": [{"name": "auctionId", "type": "uint256", "internalType": "uint256"}, {"name": "seller", "type": "address", "internalType": "address"}, {"name": "startingPrice", "type": "uint256", "internalType": "uint256"}, {"name": "endingPrice", "type": "uint256", "internalType": "uint256"}, {"name": "duration", "type": "uint256", "internalType": "uint256"}, {"name": "startedAt", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "getCurrentPrice", "type": "function", "inputs": [{"name": "_tokenId", "type": "uint256", "internalType": "uint256"}], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "getUserAuctions", "type": "function", "inputs": [{"name": "_address", "type": "address", "internalType": "address"}], "outputs": [{"name": "", "type": "uint256[]", "internalType": "uint256[]"}], "stateMutability": "view"},
    {"name": "isOnAuction", "type": "function", "inputs": [{"name": "_tokenId", "type": "uint256", "internalType": "uint256"}], "outputs": [{"name": "", "type": "bool", "internalType": "bool"}], "stateMutability": "view"},
    {"name": "jewelToken", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "address", "internalType": "contract IJewelToken"}], "stateMutability": "view"},
    {"name": "maxPrice", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "minPrice", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "owner", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "address", "internalType": "address"}], "stateMutability": "view"},
    {"name": "ownerCut", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "paused", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "bool", "internalType": "bool"}], "stateMutability": "view"},
    {"name": "renounceOwnership", "type": "function", "inputs": [], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "setFees", "type": "function", "inputs": [{"name": "_feeAddresses", "type": "address[]", "internalType": "address[]"}, {"name": "_feePercents", "type": "uint256[]", "internalType": "uint256[]"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "setLimits", "type": "function", "inputs": [{"name": "_min", "type": "uint256", "internalType": "uint256"}, {"name": "_max", "type": "uint256", "internalType": "uint256"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "transferOwnership", "type": "function", "inputs": [{"name": "newOwner", "type": "address", "internalType": "address"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "userAuctions", "type": "function", "inputs": [{"name": "", "type": "address", "internalType": "address"}, {"name": "", "type": "uint256", "internalType": "uint256"}], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"}
]
"""     

class AssistingAuction(ABIContractWrapper):
    def __init__(self, chain_key:str, rpc:str):
        contract_address = CONTRACT_ADDRESS[chain_key]
        super().__init__(contract_address=contract_address, abi=ABI, rpc=rpc)

    def auction_hero_core(self, block_identifier:BlockIdentifier = 'latest') -> address:
        return self.contract.functions.auctionHeroCore().call(block_identifier=block_identifier)

    def bid(self, cred:Credentials, _token_id:uint256, _bid_amount:uint256) -> TxReceipt:
        tx = self.contract.functions.bid(_token_id, _bid_amount)
        return self.send_transaction(tx, cred)

    def cancel_auction(self, cred:Credentials, _token_id:uint256) -> TxReceipt:
        tx = self.contract.functions.cancelAuction(_token_id)
        return self.send_transaction(tx, cred)

    def cancel_auction_when_paused(self, cred:Credentials, _token_id:uint256) -> TxReceipt:
        tx = self.contract.functions.cancelAuctionWhenPaused(_token_id)
        return self.send_transaction(tx, cred)

    def create_auction(self, cred:Credentials, _token_id:uint256, _starting_price:uint128, _ending_price:uint128, _duration:uint64, _winner:address) -> TxReceipt:
        tx = self.contract.functions.createAuction(_token_id, _starting_price, _ending_price, _duration, _winner)
        return self.send_transaction(tx, cred)

    def get_auction(self, _token_id:uint256, block_identifier:BlockIdentifier = 'latest') -> Tuple[uint256, address, uint256, uint256, uint256, uint256]:
        return self.contract.functions.getAuction(_token_id).call(block_identifier=block_identifier)

    def get_current_price(self, _token_id:uint256, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.getCurrentPrice(_token_id).call(block_identifier=block_identifier)

    def get_user_auctions(self, _address:address, block_identifier:BlockIdentifier = 'latest') -> List[uint256]:
        return self.contract.functions.getUserAuctions(_address).call(block_identifier=block_identifier)

    def is_on_auction(self, _token_id:uint256, block_identifier:BlockIdentifier = 'latest') -> bool:
        return self.contract.functions.isOnAuction(_token_id).call(block_identifier=block_identifier)

    def jewel_token(self, block_identifier:BlockIdentifier = 'latest') -> address:
        return self.contract.functions.jewelToken().call(block_identifier=block_identifier)

    def max_price(self, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.maxPrice().call(block_identifier=block_identifier)

    def min_price(self, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.minPrice().call(block_identifier=block_identifier)

    def owner(self, block_identifier:BlockIdentifier = 'latest') -> address:
        return self.contract.functions.owner().call(block_identifier=block_identifier)

    def owner_cut(self, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.ownerCut().call(block_identifier=block_identifier)

    def paused(self, block_identifier:BlockIdentifier = 'latest') -> bool:
        return self.contract.functions.paused().call(block_identifier=block_identifier)

    def renounce_ownership(self, cred:Credentials) -> TxReceipt:
        tx = self.contract.functions.renounceOwnership()
        return self.send_transaction(tx, cred)

    def set_fees(self, cred:Credentials, _fee_addresses:Sequence[address], _fee_percents:Sequence[uint256]) -> TxReceipt:
        tx = self.contract.functions.setFees(_fee_addresses, _fee_percents)
        return self.send_transaction(tx, cred)

    def set_limits(self, cred:Credentials, _min:uint256, _max:uint256) -> TxReceipt:
        tx = self.contract.functions.setLimits(_min, _max)
        return self.send_transaction(tx, cred)

    def transfer_ownership(self, cred:Credentials, new_owner:address) -> TxReceipt:
        tx = self.contract.functions.transferOwnership(new_owner)
        return self.send_transaction(tx, cred)

    def user_auctions(self, a:address, b:uint256, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.userAuctions(a, b).call(block_identifier=block_identifier)



from ..abi_contract_wrapper import ABIContractWrapper
from ..solidity_types import *
from ..credentials import Credentials

CONTRACT_ADDRESS =     {
    "cv": "0xAC9AFb5900C8A27B766bCad3A37423DC0F4C22d3",
    "sd": "0x6362b205b539afb5FC369277365441c1dC6fAa28"
}

ABI = """[
    {"name": "Approval", "type": "event", "inputs": [{"name": "owner", "type": "address", "indexed": true, "internalType": "address"}, {"name": "approved", "type": "address", "indexed": true, "internalType": "address"}, {"name": "tokenId", "type": "uint256", "indexed": true, "internalType": "uint256"}], "anonymous": false},
    {"name": "ApprovalForAll", "type": "event", "inputs": [{"name": "owner", "type": "address", "indexed": true, "internalType": "address"}, {"name": "operator", "type": "address", "indexed": true, "internalType": "address"}, {"name": "approved", "type": "bool", "indexed": false, "internalType": "bool"}], "anonymous": false},
    {"name": "Initialized", "type": "event", "inputs": [{"name": "version", "type": "uint8", "indexed": false, "internalType": "uint8"}], "anonymous": false},
    {"name": "Paused", "type": "event", "inputs": [{"name": "account", "type": "address", "indexed": false, "internalType": "address"}], "anonymous": false},
    {"name": "PetHatched", "type": "event", "inputs": [{"name": "owner", "type": "address", "indexed": true, "internalType": "address"}, {"name": "petId", "type": "uint256", "indexed": true, "internalType": "uint256"}, {"name": "pet", "type": "tuple", "components": [{"name": "id", "type": "uint256", "internalType": "uint256"}, {"name": "originId", "type": "uint8", "internalType": "uint8"}, {"name": "name", "type": "string", "internalType": "string"}, {"name": "season", "type": "uint8", "internalType": "uint8"}, {"name": "eggType", "type": "uint8", "internalType": "uint8"}, {"name": "rarity", "type": "uint8", "internalType": "uint8"}, {"name": "element", "type": "uint8", "internalType": "uint8"}, {"name": "bonusCount", "type": "uint8", "internalType": "uint8"}, {"name": "profBonus", "type": "uint8", "internalType": "uint8"}, {"name": "profBonusScalar", "type": "uint8", "internalType": "uint8"}, {"name": "craftBonus", "type": "uint8", "internalType": "uint8"}, {"name": "craftBonusScalar", "type": "uint8", "internalType": "uint8"}, {"name": "combatBonus", "type": "uint8", "internalType": "uint8"}, {"name": "combatBonusScalar", "type": "uint8", "internalType": "uint8"}, {"name": "appearance", "type": "uint16", "internalType": "uint16"}, {"name": "background", "type": "uint8", "internalType": "uint8"}, {"name": "shiny", "type": "uint8", "internalType": "uint8"}, {"name": "hungryAt", "type": "uint64", "internalType": "uint64"}, {"name": "equippableAt", "type": "uint64", "internalType": "uint64"}, {"name": "equippedTo", "type": "uint256", "internalType": "uint256"}], "indexed": false, "internalType": "struct Pet"}], "anonymous": false},
    {"name": "PetUpdated", "type": "event", "inputs": [{"name": "owner", "type": "address", "indexed": true, "internalType": "address"}, {"name": "petId", "type": "uint256", "indexed": true, "internalType": "uint256"}, {"name": "pet", "type": "tuple", "components": [{"name": "id", "type": "uint256", "internalType": "uint256"}, {"name": "originId", "type": "uint8", "internalType": "uint8"}, {"name": "name", "type": "string", "internalType": "string"}, {"name": "season", "type": "uint8", "internalType": "uint8"}, {"name": "eggType", "type": "uint8", "internalType": "uint8"}, {"name": "rarity", "type": "uint8", "internalType": "uint8"}, {"name": "element", "type": "uint8", "internalType": "uint8"}, {"name": "bonusCount", "type": "uint8", "internalType": "uint8"}, {"name": "profBonus", "type": "uint8", "internalType": "uint8"}, {"name": "profBonusScalar", "type": "uint8", "internalType": "uint8"}, {"name": "craftBonus", "type": "uint8", "internalType": "uint8"}, {"name": "craftBonusScalar", "type": "uint8", "internalType": "uint8"}, {"name": "combatBonus", "type": "uint8", "internalType": "uint8"}, {"name": "combatBonusScalar", "type": "uint8", "internalType": "uint8"}, {"name": "appearance", "type": "uint16", "internalType": "uint16"}, {"name": "background", "type": "uint8", "internalType": "uint8"}, {"name": "shiny", "type": "uint8", "internalType": "uint8"}, {"name": "hungryAt", "type": "uint64", "internalType": "uint64"}, {"name": "equippableAt", "type": "uint64", "internalType": "uint64"}, {"name": "equippedTo", "type": "uint256", "internalType": "uint256"}], "indexed": false, "internalType": "struct Pet"}], "anonymous": false},
    {"name": "Transfer", "type": "event", "inputs": [{"name": "from", "type": "address", "indexed": true, "internalType": "address"}, {"name": "to", "type": "address", "indexed": true, "internalType": "address"}, {"name": "tokenId", "type": "uint256", "indexed": true, "internalType": "uint256"}], "anonymous": false},
    {"name": "Unpaused", "type": "event", "inputs": [{"name": "account", "type": "address", "indexed": false, "internalType": "address"}], "anonymous": false},
    {"name": "approve", "type": "function", "inputs": [{"name": "to", "type": "address", "internalType": "address"}, {"name": "tokenId", "type": "uint256", "internalType": "uint256"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "balanceOf", "type": "function", "inputs": [{"name": "owner", "type": "address", "internalType": "address"}], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "bridgeMint", "type": "function", "inputs": [{"name": "_id", "type": "uint256", "internalType": "uint256"}, {"name": "_to", "type": "address", "internalType": "address"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "getApproved", "type": "function", "inputs": [{"name": "tokenId", "type": "uint256", "internalType": "uint256"}], "outputs": [{"name": "", "type": "address", "internalType": "address"}], "stateMutability": "view"},
    {"name": "getPet", "type": "function", "inputs": [{"name": "_id", "type": "uint256", "internalType": "uint256"}], "outputs": [{"name": "", "type": "tuple", "components": [{"name": "id", "type": "uint256", "internalType": "uint256"}, {"name": "originId", "type": "uint8", "internalType": "uint8"}, {"name": "name", "type": "string", "internalType": "string"}, {"name": "season", "type": "uint8", "internalType": "uint8"}, {"name": "eggType", "type": "uint8", "internalType": "uint8"}, {"name": "rarity", "type": "uint8", "internalType": "uint8"}, {"name": "element", "type": "uint8", "internalType": "uint8"}, {"name": "bonusCount", "type": "uint8", "internalType": "uint8"}, {"name": "profBonus", "type": "uint8", "internalType": "uint8"}, {"name": "profBonusScalar", "type": "uint8", "internalType": "uint8"}, {"name": "craftBonus", "type": "uint8", "internalType": "uint8"}, {"name": "craftBonusScalar", "type": "uint8", "internalType": "uint8"}, {"name": "combatBonus", "type": "uint8", "internalType": "uint8"}, {"name": "combatBonusScalar", "type": "uint8", "internalType": "uint8"}, {"name": "appearance", "type": "uint16", "internalType": "uint16"}, {"name": "background", "type": "uint8", "internalType": "uint8"}, {"name": "shiny", "type": "uint8", "internalType": "uint8"}, {"name": "hungryAt", "type": "uint64", "internalType": "uint64"}, {"name": "equippableAt", "type": "uint64", "internalType": "uint64"}, {"name": "equippedTo", "type": "uint256", "internalType": "uint256"}], "internalType": "struct Pet"}], "stateMutability": "view"},
    {"name": "getUserPets", "type": "function", "inputs": [{"name": "_address", "type": "address", "internalType": "address"}], "outputs": [{"name": "", "type": "tuple[]", "components": [{"name": "id", "type": "uint256", "internalType": "uint256"}, {"name": "originId", "type": "uint8", "internalType": "uint8"}, {"name": "name", "type": "string", "internalType": "string"}, {"name": "season", "type": "uint8", "internalType": "uint8"}, {"name": "eggType", "type": "uint8", "internalType": "uint8"}, {"name": "rarity", "type": "uint8", "internalType": "uint8"}, {"name": "element", "type": "uint8", "internalType": "uint8"}, {"name": "bonusCount", "type": "uint8", "internalType": "uint8"}, {"name": "profBonus", "type": "uint8", "internalType": "uint8"}, {"name": "profBonusScalar", "type": "uint8", "internalType": "uint8"}, {"name": "craftBonus", "type": "uint8", "internalType": "uint8"}, {"name": "craftBonusScalar", "type": "uint8", "internalType": "uint8"}, {"name": "combatBonus", "type": "uint8", "internalType": "uint8"}, {"name": "combatBonusScalar", "type": "uint8", "internalType": "uint8"}, {"name": "appearance", "type": "uint16", "internalType": "uint16"}, {"name": "background", "type": "uint8", "internalType": "uint8"}, {"name": "shiny", "type": "uint8", "internalType": "uint8"}, {"name": "hungryAt", "type": "uint64", "internalType": "uint64"}, {"name": "equippableAt", "type": "uint64", "internalType": "uint64"}, {"name": "equippedTo", "type": "uint256", "internalType": "uint256"}], "internalType": "struct Pet[]"}], "stateMutability": "view"},
    {"name": "hatchPet", "type": "function", "inputs": [{"name": "_petOptions", "type": "tuple", "components": [{"name": "originId", "type": "uint8", "internalType": "uint8"}, {"name": "name", "type": "string", "internalType": "string"}, {"name": "season", "type": "uint8", "internalType": "uint8"}, {"name": "eggType", "type": "uint8", "internalType": "uint8"}, {"name": "rarity", "type": "uint8", "internalType": "uint8"}, {"name": "element", "type": "uint8", "internalType": "uint8"}, {"name": "bonusCount", "type": "uint8", "internalType": "uint8"}, {"name": "profBonus", "type": "uint8", "internalType": "uint8"}, {"name": "profBonusScalar", "type": "uint8", "internalType": "uint8"}, {"name": "craftBonus", "type": "uint8", "internalType": "uint8"}, {"name": "craftBonusScalar", "type": "uint8", "internalType": "uint8"}, {"name": "combatBonus", "type": "uint8", "internalType": "uint8"}, {"name": "combatBonusScalar", "type": "uint8", "internalType": "uint8"}, {"name": "appearance", "type": "uint16", "internalType": "uint16"}, {"name": "background", "type": "uint8", "internalType": "uint8"}, {"name": "shiny", "type": "uint8", "internalType": "uint8"}], "internalType": "struct PetOptions"}, {"name": "owner", "type": "address", "internalType": "address"}], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "nonpayable"},
    {"name": "initialize", "type": "function", "inputs": [{"name": "_name", "type": "string", "internalType": "string"}, {"name": "_symbol", "type": "string", "internalType": "string"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "isApprovedForAll", "type": "function", "inputs": [{"name": "owner", "type": "address", "internalType": "address"}, {"name": "operator", "type": "address", "internalType": "address"}], "outputs": [{"name": "", "type": "bool", "internalType": "bool"}], "stateMutability": "view"},
    {"name": "name", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "string", "internalType": "string"}], "stateMutability": "view"},
    {"name": "nextPetId", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "ownerOf", "type": "function", "inputs": [{"name": "tokenId", "type": "uint256", "internalType": "uint256"}], "outputs": [{"name": "", "type": "address", "internalType": "address"}], "stateMutability": "view"},
    {"name": "pause", "type": "function", "inputs": [], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "paused", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "bool", "internalType": "bool"}], "stateMutability": "view"},
    {"name": "safeTransferFrom", "type": "function", "inputs": [{"name": "from", "type": "address", "internalType": "address"}, {"name": "to", "type": "address", "internalType": "address"}, {"name": "tokenId", "type": "uint256", "internalType": "uint256"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "safeTransferFrom", "type": "function", "inputs": [{"name": "from", "type": "address", "internalType": "address"}, {"name": "to", "type": "address", "internalType": "address"}, {"name": "tokenId", "type": "uint256", "internalType": "uint256"}, {"name": "_data", "type": "bytes", "internalType": "bytes"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "setApprovalForAll", "type": "function", "inputs": [{"name": "operator", "type": "address", "internalType": "address"}, {"name": "approved", "type": "bool", "internalType": "bool"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "symbol", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "string", "internalType": "string"}], "stateMutability": "view"},
    {"name": "tokenByIndex", "type": "function", "inputs": [{"name": "index", "type": "uint256", "internalType": "uint256"}], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "tokenOfOwnerByIndex", "type": "function", "inputs": [{"name": "owner", "type": "address", "internalType": "address"}, {"name": "index", "type": "uint256", "internalType": "uint256"}], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "tokenURI", "type": "function", "inputs": [{"name": "tokenId", "type": "uint256", "internalType": "uint256"}], "outputs": [{"name": "", "type": "string", "internalType": "string"}], "stateMutability": "view"},
    {"name": "totalSupply", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "transferFrom", "type": "function", "inputs": [{"name": "from", "type": "address", "internalType": "address"}, {"name": "to", "type": "address", "internalType": "address"}, {"name": "tokenId", "type": "uint256", "internalType": "uint256"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "unpause", "type": "function", "inputs": [], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "updatePet", "type": "function", "inputs": [{"name": "_pet", "type": "tuple", "components": [{"name": "id", "type": "uint256", "internalType": "uint256"}, {"name": "originId", "type": "uint8", "internalType": "uint8"}, {"name": "name", "type": "string", "internalType": "string"}, {"name": "season", "type": "uint8", "internalType": "uint8"}, {"name": "eggType", "type": "uint8", "internalType": "uint8"}, {"name": "rarity", "type": "uint8", "internalType": "uint8"}, {"name": "element", "type": "uint8", "internalType": "uint8"}, {"name": "bonusCount", "type": "uint8", "internalType": "uint8"}, {"name": "profBonus", "type": "uint8", "internalType": "uint8"}, {"name": "profBonusScalar", "type": "uint8", "internalType": "uint8"}, {"name": "craftBonus", "type": "uint8", "internalType": "uint8"}, {"name": "craftBonusScalar", "type": "uint8", "internalType": "uint8"}, {"name": "combatBonus", "type": "uint8", "internalType": "uint8"}, {"name": "combatBonusScalar", "type": "uint8", "internalType": "uint8"}, {"name": "appearance", "type": "uint16", "internalType": "uint16"}, {"name": "background", "type": "uint8", "internalType": "uint8"}, {"name": "shiny", "type": "uint8", "internalType": "uint8"}, {"name": "hungryAt", "type": "uint64", "internalType": "uint64"}, {"name": "equippableAt", "type": "uint64", "internalType": "uint64"}, {"name": "equippedTo", "type": "uint256", "internalType": "uint256"}], "internalType": "struct Pet"}], "outputs": [], "stateMutability": "nonpayable"}
]
"""     

class PetCore(ABIContractWrapper):
    def __init__(self, chain_key:str, rpc:str):
        contract_address = CONTRACT_ADDRESS[chain_key]
        super().__init__(contract_address=contract_address, abi=ABI, rpc=rpc)

    def approve(self, cred:Credentials, to:address, token_id:uint256) -> TxReceipt:
        tx = self.contract.functions.approve(to, token_id)
        return self.send_transaction(tx, cred)

    def balance_of(self, owner:address, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.balanceOf(owner).call(block_identifier=block_identifier)

    def bridge_mint(self, cred:Credentials, _id:uint256, _to:address) -> TxReceipt:
        tx = self.contract.functions.bridgeMint(_id, _to)
        return self.send_transaction(tx, cred)

    def get_approved(self, token_id:uint256, block_identifier:BlockIdentifier = 'latest') -> address:
        return self.contract.functions.getApproved(token_id).call(block_identifier=block_identifier)

    def get_pet(self, _id:uint256, block_identifier:BlockIdentifier = 'latest') -> tuple:
        return self.contract.functions.getPet(_id).call(block_identifier=block_identifier)

    def get_user_pets(self, _address:address, block_identifier:BlockIdentifier = 'latest') -> List[tuple]:
        return self.contract.functions.getUserPets(_address).call(block_identifier=block_identifier)

    def hatch_pet(self, cred:Credentials, _pet_options:tuple, owner:address) -> TxReceipt:
        tx = self.contract.functions.hatchPet(_pet_options, owner)
        return self.send_transaction(tx, cred)

    def initialize(self, cred:Credentials, _name:string, _symbol:string) -> TxReceipt:
        tx = self.contract.functions.initialize(_name, _symbol)
        return self.send_transaction(tx, cred)

    def is_approved_for_all(self, owner:address, operator:address, block_identifier:BlockIdentifier = 'latest') -> bool:
        return self.contract.functions.isApprovedForAll(owner, operator).call(block_identifier=block_identifier)

    def name(self, block_identifier:BlockIdentifier = 'latest') -> string:
        return self.contract.functions.name().call(block_identifier=block_identifier)

    def next_pet_id(self, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.nextPetId().call(block_identifier=block_identifier)

    def owner_of(self, token_id:uint256, block_identifier:BlockIdentifier = 'latest') -> address:
        return self.contract.functions.ownerOf(token_id).call(block_identifier=block_identifier)

    def pause(self, cred:Credentials) -> TxReceipt:
        tx = self.contract.functions.pause()
        return self.send_transaction(tx, cred)

    def paused(self, block_identifier:BlockIdentifier = 'latest') -> bool:
        return self.contract.functions.paused().call(block_identifier=block_identifier)

    def safe_transfer_from(self, cred:Credentials, _from:address, to:address, token_id:uint256) -> TxReceipt:
        tx = self.contract.functions.safeTransferFrom(_from, to, token_id)
        return self.send_transaction(tx, cred)

    def safe_transfer_from(self, cred:Credentials, _from:address, to:address, token_id:uint256, _data:bytes) -> TxReceipt:
        tx = self.contract.functions.safeTransferFrom(_from, to, token_id, _data)
        return self.send_transaction(tx, cred)

    def set_approval_for_all(self, cred:Credentials, operator:address, approved:bool) -> TxReceipt:
        tx = self.contract.functions.setApprovalForAll(operator, approved)
        return self.send_transaction(tx, cred)

    def symbol(self, block_identifier:BlockIdentifier = 'latest') -> string:
        return self.contract.functions.symbol().call(block_identifier=block_identifier)

    def token_by_index(self, index:uint256, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.tokenByIndex(index).call(block_identifier=block_identifier)

    def token_of_owner_by_index(self, owner:address, index:uint256, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.tokenOfOwnerByIndex(owner, index).call(block_identifier=block_identifier)

    def token_uri(self, token_id:uint256, block_identifier:BlockIdentifier = 'latest') -> string:
        return self.contract.functions.tokenURI(token_id).call(block_identifier=block_identifier)

    def total_supply(self, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.totalSupply().call(block_identifier=block_identifier)

    def transfer_from(self, cred:Credentials, _from:address, to:address, token_id:uint256) -> TxReceipt:
        tx = self.contract.functions.transferFrom(_from, to, token_id)
        return self.send_transaction(tx, cred)

    def unpause(self, cred:Credentials) -> TxReceipt:
        tx = self.contract.functions.unpause()
        return self.send_transaction(tx, cred)

    def update_pet(self, cred:Credentials, _pet:tuple) -> TxReceipt:
        tx = self.contract.functions.updatePet(_pet)
        return self.send_transaction(tx, cred)


#! /usr/bin/env python
from web3 import Web3
from web3.middleware.geth_poa import geth_poa_middleware
from web3.logs import DISCARD

from .credentials import Credentials

from .solidity_types import (address, ChecksumAddress, TxReceipt, AttributeDict)
from web3.contract.contract import Contract
from typing import Dict, Tuple, Union, Optional, Any, Sequence

DEFAULT_TIMEOUT = 30
DEFAULT_MAX_GAS = 50
DEFAULT_MAX_PRIORITY_GAS = 3

W3_INSTANCES: Dict[str, Web3] = {}

class ABIContractWrapper:
    def __init__(self, 
                 contract_address:str, 
                 abi:str,
                 rpc:str,
                 max_gas_gwei:float=DEFAULT_MAX_GAS,
                 max_priority_gwei:float=DEFAULT_MAX_PRIORITY_GAS):
        self.rpc = rpc
        # self.contract_address = contract_address
        self.abi = abi
        self.nonces: Dict[address, int] = {}
        self.timeout = DEFAULT_TIMEOUT

        # This one superclass may be used by many contracts, who don't all need to
        # create separate Web3 instances. Just use one per RPC
        global W3_INSTANCES
        w3 = W3_INSTANCES.get(self.rpc, None)
        if not w3:
            w3 = Web3(Web3.HTTPProvider(self.rpc))
            w3.middleware_onion.inject(geth_poa_middleware, layer=0)
            W3_INSTANCES[self.rpc] = w3
        self.w3 = w3
        self.contract_address:ChecksumAddress = self.w3.to_checksum_address(contract_address)

        self.max_gas_wei = self.w3.to_wei(max_gas_gwei, 'gwei')
        self.max_priority_wei = self.w3.to_wei(max_priority_gwei, 'gwei')

        self.contract = self.w3.eth.contract(self.contract_address, abi=self.abi)

    def get_nonce_and_update(self, address:address, force_fetch=True) -> int:
        # FIXME: I think there's a bug in the caching logic below that lets
        # nonces run ahead of where they should be and causes transactions to
        # fail.
        # So force_fetch is set to true, which fetches the appropriate nonce for
        # every transaction
        # - Athiriyya 13 December 2022

        # We keep track of our own nonce, and only re-fetch it if a 'nonce too low'
        # error gets thrown       
        nonce = self.nonces.get(address, 0) 
        if force_fetch or nonce == 0:
            nonce = self.w3.eth.get_transaction_count( address, 'pending')

        # Store the next nonce this address will use, and return the current one
        self.nonces[address] = nonce + 1
        return nonce

    def get_gas_dict_and_update(self, address:address) -> Dict[str, int]:
        nonce = self.get_nonce_and_update(address)
         
        legacy = False
        if legacy:
            # TODO: it's expensive to query for fees with every transaction. 
            # Maybe query only once a minute?
            gas, gas_price = self.get_legacy_gas_fee()
            gas_dict = {
                'gas': gas, 
                'gasPrice':gas_price, 
                'nonce':nonce
            }
        else:
            gas_dict = {
                'from': address, 
                'maxFeePerGas': int(self.max_gas_wei), 
                'maxPriorityFeePerGas': int(self.max_priority_wei), 
                'nonce': nonce
            }
        return gas_dict

    def call_contract_function(self, function_name:str, *args) -> Any:
        contract_func = getattr(self.contract, function_name)
        return contract_func(*args).call()

    def get_custom_contract(self, contract_address:ChecksumAddress, abi:str | None=None) -> Contract:
        # TODO: Many custom contracts for e.g. ERC20 tokens could
        # be re-used by caching a contracts dictionary keyed by address
        # For now, just return a new contract
        abi = abi or self.abi
        checked_addr = self.w3.to_checksum_address(contract_address)
        contract = self.w3.eth.contract(checked_addr, abi=abi)
        return contract

    def send_transaction(self,
                         tx,
                         cred:Credentials,
                         extra_dict:Dict[str,Any] | None = None
                        ) -> TxReceipt:
        # Some transactions require extra information or fees when building 
        # the transaction. e.g. bridging functions need a {'value': <bridge_fee_in_wei>}
        # argument. If supplied, add that extra info
        address = cred.address  
        gas_dict = self.get_gas_dict_and_update(address)
        if extra_dict:
            gas_dict.update(extra_dict)
        tx_dict = tx.build_transaction(gas_dict)    
        signed_tx = self.w3.eth.account.sign_transaction(tx_dict, private_key=cred.private_key)
        try:
            self.w3.eth.send_raw_transaction(signed_tx.rawTransaction)

        except Exception as e:
            if 'nonce too low' in str(e):
                nonce = self.get_nonce_and_update(address, force_fetch=True)
                return self.send_transaction( tx, cred)
            # otherwise, raise
            raise(e)

        receipt = self.w3.eth.wait_for_transaction_receipt(
            transaction_hash=signed_tx.hash,
            poll_latency=1,
            timeout=self.timeout,
        )
        return receipt

    def get_legacy_gas_fee(self) ->Tuple[int, int]:
        # See: https://web3py.readthedocs.io/en/stable/gas_price.html#gas-price-api
        # Some transactions may require a gas dict with the keys {'gasPrice': x_wei, '': y_wei}
        block = self.w3.eth.get_block("pending")
        base_gas = block.get('gasUsed', 2_500_000) + self.w3.to_wei(50, 'gwei')
        gas_limit =block.get('gasLimit', 2_500_000)

        return base_gas, gas_limit

    def tx_receipt_for_hash(self, tx_hash:address) -> TxReceipt:
        tx_receipt = self.w3.eth.get_transaction_receipt(tx_hash)
        return tx_receipt

    def parse_events(self, tx_receipt:TxReceipt, event_names:Sequence[str] | None = None) -> Dict[str, AttributeDict]:
        event_dicts = {}
        for event in self.contract.events: # type: ignore
            eds = event().process_receipt(tx_receipt, errors=DISCARD)
            if eds:
                for ed in eds:
                    event_dicts.setdefault(ed.event,[]).append(ed)
        if event_names:
            event_dicts = {k:v for k,v in event_dicts.items() if k in event_names}
        return event_dicts        
